local VERSION = tonumber(("$Revision: 44490 $"):match("%d+"))

local Parrot = Parrot
local Parrot_AnimationStyles = Parrot:NewModule("AnimationStyles")
if Parrot.revision < VERSION then
	Parrot.version = "r" .. VERSION
	Parrot.revision = VERSION
	Parrot.date = ("$Date: 2007-07-22 03:52:24 -0400 (Sun, 22 Jul 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")
end

local L = Parrot:L("Parrot_AnimationStyles")

local animationStyles = {}
local animationStylesValidate = {}

--[[----------------------------------------------------------------------------------
Arguments:
	table - a data table holding the details of an animation style.
Notes:
	* The data table is of the following style:
	<pre>{
		name = "Name of the style in English",
		localName = "Name of the style in the current locale",
		func = function(
			fontstring, -- the FontString to work with
			xOffset, -- the horizontal offset from the center of the screen
			yOffset, -- the vertical offset from the center of the screen
			size, -- the size of the animation area to work within.
			percent, -- [0, 1] the current percentage done.
			direction, -- a string of the direction specified.
			num, -- the number of the FontString out of the others using the same area.
			max, -- the number of FontStrings using the same area
			uid, -- the unique ID of the FontString, for the given area.
		)
			fontstring:SetPoint(...) -- the FontString must be positioned here.
		end,
		init = function( -- this is optional
			fontstring, -- the FontString to work with
			xOffset, -- the horizontal offset from the center of the screen
			yOffset, -- the vertical offset from the center of the screen
			size, -- the size of the animation area to work within.
			direction, -- a string of the direction specified.
			uid, -- the unique ID of the FontString, for the given area.
		)
			-- this is called before the first func() is called.
		end,
		cleanup = function( -- this is optional
			fontstring, -- the FontString to work with
			xOffset, -- the horizontal offset from the center of the screen
			yOffset, -- the vertical offset from the center of the screen
			size, -- the size of the animation area to work within.
			direction, -- a string of the direction specified.
			uid, -- the unique ID of the FontString, for the given area.
		)
			-- this is called after the last func() is called.
		end,
		overlap = true, -- or false. This dictates whether FontStrings should be moved along prematurely if they are overlapping.
		defaultDirection = "UP", -- required if directions is given.
		directions = { -- this is optional
			UP = "Up" -- key-value of backend direction to localized name. This will be exposed to the user, who can choose what direction to go with.
		}
	}</pre>
	* the following fields are available on the fontString: .font, .fontSize, and .fontOutline. With these, you can manipulate the size of the fontstring with <tt>fs:SetFont(fs.font, fs.fontSize*2, fs.fontOutline)</tt> to double its size.
Example:
	Parrot_AnimationStyles:RegisterAnimationStyle {
		-- simple vertical
		name = "Straight", 
		localName = L["Straight"],
		func = function(fontstring, xOffset, yOffset, size, percent, direction)
			if direction == "DOWN" then
				percent = 1 - percent
			end
			local y = yOffset + (percent - 0.5) * size
			local x = xOffset
			fontstring:SetPoint("CENTER", UIParent, "CENTER", x, y)
		end,
		defaultDirection = "DOWN",
		directions = {
			UP = L["Up"],
			DOWN = L["Down"],
		},
		overlap = true,
	}
------------------------------------------------------------------------------------]]
function Parrot_AnimationStyles:RegisterAnimationStyle(data)
	self = Parrot_AnimationStyles -- so people can Parrot:RegisterAnimationStyle
	if type(data) ~= "table" then
		error(("Bad argument #2 to `RegisterAnimationStyle'. Expected %q, got %q."):format("table", type(data)), 2)
	end
	local name = data.name
	if type(name) ~= "string" then
		error(("Bad argument #2 to `RegisterAnimationStyle'. name must be a %q, got %q."):format("string", type(name)), 2)
	end
	local localName = data.localName
	if type(localName) ~= "string" then
		error(("Bad argument #2 to `RegisterAnimationStyle'. localName must be a %q, got %q."):format("string", type(localName)), 2)
	end
	if type(data.func) ~= "function" then
		error(("Bad argument #2 to `RegisterAnimationStyle'. func must be a %q, got %q."):format("function", type(data.func)), 2)
	end
	if data.cleanup and type(data.cleanup) ~= "function" then
		error(("Bad argument #2 to `RegisterAnimationStyle'. cleanup must be a %q or %q, got %q."):format("function", "nil", type(data.cleanup)), 2)
	end
	if data.init and type(data.init) ~= "function" then
		error(("Bad argument #2 to `RegisterAnimationStyle'. init must be a %q or %q, got %q."):format("function", "nil", type(data.init)), 2)
	end
	if data.overlap and data.overlap ~= true then
		error(("Bad argument #2 to `RegisterAnimationStyle'. overlap must be a %q, got %q."):format("boolean", type(data.overlap)), 2)
	end
	if data.directions then
		if type(data.directions) ~= "table" then
			error(("Bad argument #2 to `RegisterAnimationStyle'. directions must be a %q, got %q."):format("table", type(data.directions)), 2)
		end
		if type(data.defaultDirection) ~= "string" then
			error(("Bad argument #2 to `RegisterAnimationStyle'. defaultDirection must be a %q, got %q."):format("string", type(data.defaultDirection)), 2)
		end
	end
	if animationStyles[name] then
		error(("Animation style %q already registered"):format(name), 2)
	end
	animationStyles[name] = data
	animationStylesValidate[name] = localName
end
Parrot.RegisterAnimationStyle = Parrot_AnimationStyles.RegisterAnimationStyle

--[[----------------------------------------------------------------------------------
Arguments:
	string - the name of the animation style, in English.
Returns:
	boolean - whether the animation style has been registered or not.
Example:
	local has = Parrot:HasAnimationStyle("My funky style")
------------------------------------------------------------------------------------]]
function Parrot_AnimationStyles:HasAnimationStyle(name)
	self = Parrot_AnimationStyles -- so people can Parrot:RegisterAnimationStyle
	AceLibrary.argCheck(self, name, 2, "string")
	return not not animationStyles[name]
end
Parrot.HasAnimationStyle = Parrot_AnimationStyles.HasAnimationStyle

-- #NODOC
function Parrot_AnimationStyles:GetAnimationStyle(name)
	return animationStyles[name] or animationStyles.Straight
end

-- #NODOC
function Parrot_AnimationStyles:GetAnimationStylesValidate()
	return animationStylesValidate
end

-- #NODOC
function Parrot_AnimationStyles:GetAnimationStyleDirectionValidate(name)
	if not animationStyles[name] then
		return
	end
	return animationStyles[name].directions
end
function Parrot_AnimationStyles:GetAnimationStyleDefaultDirection(name)
	if not animationStyles[name] then
		return nil
	end
	return animationStyles[name].defaultDirection
end

------------------------------------------------------------------------------
-- Animation style registry --------------------------------------------------
------------------------------------------------------------------------------

Parrot_AnimationStyles:RegisterAnimationStyle {
	-- simple vertical
	name = "Straight", 
	localName = L["Straight"],
	func = function(fontstring, xOffset, yOffset, size, percent, direction)
		if direction == "DOWN" then
			percent = 1 - percent
		end
		local y = yOffset + (percent - 0.5) * size
		local x = xOffset
		fontstring:SetPoint("CENTER", UIParent, "CENTER", x, y)
	end,
	defaultDirection = "DOWN",
	directions = {
		UP = L["Up"],
		DOWN = L["Down"],
	},
	overlap = true,
}
Parrot_AnimationStyles:RegisterAnimationStyle {
	-- makes a parabola in the form of x = y^2 - 1
	name = "Parabola",
	localName = L["Parabola"],
	func = function(fontstring, xOffset, yOffset, size, percent, direction, num, max, uid)
		local vert, horiz = (";"):split(direction)
		if vert == "DOWN" then
			percent = 1 - percent
		end
		local xDiff = (1 - 4*(percent - 0.5)^2) * size * 1/3
		local point = "LEFT"
		if horiz == "LEFT" or (horiz == "ALT" and uid%2 == 0) then
			xDiff = -xDiff
			point = "RIGHT"
		end
		local y = yOffset + (percent - 0.5) * size
		local x = xOffset + xDiff
		fontstring:SetPoint(point, UIParent, "CENTER", x, y)
	end,
	defaultDirection = "DOWN;ALT",
	directions = {
		["UP;LEFT"] = L["Up, left"],
		["UP;RIGHT"] = L["Up, right"],
		["UP;ALT"] = L["Up, alternating"],
		["DOWN;LEFT"] = L["Down, left"],
		["DOWN;RIGHT"] = L["Down, right"],
		["DOWN;ALT"] = L["Down, alternating"],
	},
	overlap = true,
}
Parrot_AnimationStyles:RegisterAnimationStyle {
	-- makes a semicircle in the form of x = -(-y^2 + 1)^0.5
	name = "Semicircle",
	localName = L["Semicircle"],
	func = function(fontstring, xOffset, yOffset, size, percent, direction, num, max, uid)
		local vert, horiz = (";"):split(direction)
		local yDiff = size/2 * sin(90 + 180 * (percent))
		local xDiff = -size/2 * cos(90 + 180 * (percent))
		if vert == "UP" then
			yDiff = -yDiff
		end
		local point = "LEFT"
		if horiz == "LEFT" or (horiz == "ALT" and uid%2 == 0) then
			xDiff = -xDiff
			point = "RIGHT"
		end
		local y = yOffset + yDiff
		local x = xOffset + xDiff
		fontstring:SetPoint(point, UIParent, "CENTER", x, y)
	end,
	defaultDirection = "DOWN;ALT",
	directions = {
		["UP;LEFT"] = L["Up, left"],
		["UP;RIGHT"] = L["Up, right"],
		["UP;ALT"] = L["Up, alternating"],
		["DOWN;LEFT"] = L["Down, left"],
		["DOWN;RIGHT"] = L["Down, right"],
		["DOWN;ALT"] = L["Down, alternating"],
	},
	overlap = true
}

Parrot_AnimationStyles:RegisterAnimationStyle {
	-- places in the center and shakes around
	name = "Pow",
	localName = L["Pow"],
	func = function(fontstring, xOffset, yOffset, size, percent, direction, num, max, uid)
		local fsHeight = fontstring:GetHeight()
		local topDiff = (max - 1) * 0.5 * fsHeight
		local currDiff = -(num - 1) * fsHeight
		local yDiff = topDiff + currDiff
		if direction == "DOWN" then
			yDiff = -yDiff
		end
		local y = yOffset + yDiff
		local x = xOffset
		-- now for squiggle
		local now = GetTime()
		if not fontstring.nextSquiggle or fontstring.nextSquiggle < now then
			fontstring.nextSquiggle = now + 0.05
			fontstring.squiggleX = math.random(-1, 1)
			fontstring.squiggleY = math.random(-1, 1)
		end	
		x = x + fontstring.squiggleX
		y = y + fontstring.squiggleY
	
		fontstring:SetPoint("CENTER", UIParent, "CENTER", x, y)
	end,
	cleanup = function(fontstring, scrollArea)
		fontstring.nextSquiggle = nil
		fontstring.squiggleX = nil
		fontstring.squiggleY = nil
	end,
	defaultDirection = "DOWN",
	directions = {
		UP = L["Up"],
		DOWN = L["Down"],
	},
}

Parrot_AnimationStyles:RegisterAnimationStyle {
	-- places in the center
	name = "Static",
	localName = L["Static"],
	func = function(fontstring, xOffset, yOffset, size, percent, direction, num, max, uid)
		local fsHeight = fontstring:GetHeight()
		local topDiff = (max - 1) * 0.5 * fsHeight
		local currDiff = -(num - 1) * fsHeight
		local yDiff = topDiff + currDiff
		if direction == "DOWN" then
			yDiff = -yDiff
		end
		local y = yOffset + yDiff
		local x = xOffset
		fontstring:SetPoint("CENTER", UIParent, "CENTER", x, y)
	end,
	defaultDirection = "DOWN",
	directions = {
		UP = L["Up"],
		DOWN = L["Down"],
	},
}

Parrot_AnimationStyles:RegisterAnimationStyle({
	-- makes a parabola in the form of y = (-x - 1)(-x + 0.5) + 1
	name = "Rainbow",
	localName = L["Rainbow"],
	init = function(fontstring, xOffset, yOffset, size, direction, uid)
		fontstring.leftRoot = 1 + (math.random()-0.5) / 5
		fontstring.rightRoot = -0.5 + (math.random()-0.5) / 5
	end,
	func = function(fontstring, xOffset, yOffset, size, percent, direction, num, max, uid)
		local xDiff = size/2 * percent
		local yDiff = size * (-16/9 * (percent - fontstring.leftRoot) * (percent - fontstring.rightRoot) - 0.5)
		local vert, horiz = (";"):split(direction)
		if vert == "UP" then
			yDiff = -yDiff
		end
		local point = "LEFT"
		if horiz == "LEFT" or (horiz == "ALT" and uid%2 == 0) then
			xDiff = -xDiff
			point = "RIGHT"
		end
		local y = yOffset + yDiff
		local x = xOffset + xDiff
		fontstring:SetPoint(point, UIParent, "CENTER", x, y)
	end,
	cleanup = function(fontstring, xOffset, yOffset, size, direction, uid)
		fontstring.leftRoot = nil
		fontstring.rightRoot = nil
	end,
	overlap = false,
	defaultDirection = "DOWN;ALT",
	directions = {
		["UP;LEFT"] = L["Up, left"],
		["UP;RIGHT"] = L["Up, right"],
		["UP;ALT"] = L["Up, alternating"],
		["DOWN;LEFT"] = L["Down, left"],
		["DOWN;RIGHT"] = L["Down, right"],
		["DOWN;ALT"] = L["Down, alternating"],
	},
})

Parrot_AnimationStyles:RegisterAnimationStyle({
	-- makes a path going straight to the right with a random y.
	name = "Horizontal",
	localName = L["Horizontal"],
	init = function(fontstring, xOffset, yOffset, size, direction, uid)
		fontstring.y = math.random() - 0.5
	end,
	func = function(fontstring, xOffset, yOffset, size, percent, direction, num, max, uid)
		local xDiff = size * percent
		local point = "LEFT"
		if direction == "LEFT" or (direction == "ALT" and uid%2 == 0) then
			xDiff = -xDiff
			point = "RIGHT"
		end
		local y = yOffset + size * fontstring.y
		local x = xOffset + xDiff
		fontstring:SetPoint(point, UIParent, "CENTER", x, y)
	end,
	cleanup = function(fontstring, xOffset, yOffset, size, direction, uid)
		fontstring.y = nil
	end,
	overlap = false,
	defaultDirection = "ALT",
	directions = {
		["LEFT"] = L["Left"],
		["RIGHT"] = L["Right"],
		["ALT"] = L["Alternating"],
	},
})


Parrot_AnimationStyles:RegisterAnimationStyle({
	-- makes a parabola in the form of y = (-x - 1)(-x + 0.5) + 1
	name = "Angled",
	localName = L["Angled"],
	init = function(fontstring, xOffset, yOffset, size, direction, uid)
		fontstring.finishX = size/2 * (math.random()/2 + 0.75)
		fontstring.finishY = size/2 * (math.random()/2 + 0.75)
	end,
	func = function(fontstring, xOffset, yOffset, size, percent, direction, num, max, uid)
		local xDiff, yDiff
		if percent < 0.3 then
			xDiff = percent/0.3 * fontstring.finishX
			yDiff = -percent/0.3 * fontstring.finishY
		elseif percent < 0.8 then
			local now = GetTime()
			if not fontstring.nextSquiggle or now > fontstring.nextSquiggle then
				fontstring.nextSquiggle = now + 0.05
				fontstring.squiggleX = math.random(-1, 1)
				fontstring.squiggleY = math.random(-1, 1)
			end
			xDiff = fontstring.finishX + fontstring.squiggleX
			yDiff = -fontstring.finishY + fontstring.squiggleY
		else
			xDiff = ((percent - 0.8)/0.2 + 1) * fontstring.finishX
			yDiff = -(1 - percent)/0.2 * fontstring.finishY
		end
		
		local vert, horiz = (";"):split(direction)
		if vert == "UP" then
			yDiff = -yDiff
		end
		local point = "LEFT"
		if horiz == "LEFT" or (horiz == "ALT" and uid%2 == 0) then
			xDiff = -xDiff
			point = "RIGHT"
		end
		local y = yOffset + yDiff
		local x = xOffset + xDiff
		fontstring:SetPoint(point, UIParent, "CENTER", x, y)
	end,
	cleanup = function(fontstring, xOffset, yOffset, size, direction, uid)
		fontstring.nextSquiggle = nil
		fontstring.squiggleX = nil
		fontstring.squiggleY = nil
		fontstring.finishX = nil
		fontstring.finishY = nil
	end,
	overlap = false,
	defaultDirection = "DOWN;ALT",
	directions = {
		["UP;LEFT"] = L["Up, left"],
		["UP;RIGHT"] = L["Up, right"],
		["UP;ALT"] = L["Up, alternating"],
		["DOWN;LEFT"] = L["Down, left"],
		["DOWN;RIGHT"] = L["Down, right"],
		["DOWN;ALT"] = L["Down, alternating"],
	},
})


Parrot_AnimationStyles:RegisterAnimationStyle({
	-- makes a parabola in the form of y = (-x - 1)(-x + 0.5) + 1
	name = "Sprinkler",
	localName = L["Sprinkler"],
	init = function(fontstring, xOffset, yOffset, size, direction, uid)
		local dir, clock = (";"):split(direction)
		local base
		if dir == "RIGHT" then
			base = 0
		elseif dir == "TOP" then
			base = 90
		elseif dir == "LEFT" then
			base = 180
		else -- bottom
			base = 270
		end
		local slices = math.floor(size / fontstring:GetHeight() / 2)
		local diff = (uid%slices)*(120/(slices - 1)) - 60
		if clock == "CCW" then
			diff = -diff
		end
		fontstring.angle = base + diff
	end,
	func = function(fontstring, xOffset, yOffset, size, percent, direction, num, max, uid)
		local xDiff, yDiff
		if percent < 0.3 then
			xDiff = percent/0.3 * size/2 * cos(fontstring.angle)
			yDiff = percent/0.3 * size/2 * sin(fontstring.angle)
		elseif percent < 0.8 then
			local now = GetTime()
			if not fontstring.nextSquiggle or now > fontstring.nextSquiggle then
				fontstring.nextSquiggle = now + 0.05
				fontstring.squiggleX = math.random(-1, 1)
				fontstring.squiggleY = math.random(-1, 1)
			end
			xDiff = size/2 * cos(fontstring.angle) + fontstring.squiggleX
			yDiff = size/2 * sin(fontstring.angle) + fontstring.squiggleY
		else
			xDiff = ((percent - 0.8)/0.2 + 1) * size/2 * cos(fontstring.angle)
			yDiff = ((percent - 0.8)/0.2 + 1) * size/2 * sin(fontstring.angle)
		end
		
		local y = yOffset + yDiff
		local x = xOffset + xDiff
		local dir, clock = (";"):split(direction)
		local point = "CENTER"
		if dir == "LEFT" then
			point = "RIGHT"
		elseif dir == "RIGHT" then
			point = "LEFT"
		end
		fontstring:SetPoint(point, UIParent, "CENTER", x, y)
	end,
	cleanup = function(fontstring, xOffset, yOffset, size, direction, uid)
		fontstring.nextSquiggle = nil
		fontstring.squiggleX = nil
		fontstring.squiggleY = nil
		fontstring.finishX = nil
		fontstring.finishY = nil
	end,
	overlap = false,
	defaultDirection = "UP;CCW",
	directions = {
		["UP;CW"] = L["Up, clockwise"],
		["DOWN;CW"] = L["Down, clockwise"],
		["LEFT;CW"] = L["Left, clockwise"],
		["RIGHT;CW"] = L["Right, clockwise"],
		["UP;CCW"] = L["Up, counter-clockwise"],
		["DOWN;CCW"] = L["Down, counter-clockwise"],
		["LEFT;CCW"] = L["Left, counter-clockwise"],
		["RIGHT;CCW"] = L["Right, counter-clockwise"],
	},
})
